# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import json

from django.db import models, migrations
from game.level_management import set_blocks_inner, set_decor_inner


def add_cows_block(apps, schema_editor):

    Block = apps.get_model("game", "Block")

    declare_event = Block.objects.create(type="declare_event")
    puff_up = Block.objects.create(type="puff_up")
    sound_horn = Block.objects.create(type="sound_horn")


class Migration(migrations.Migration):

    dependencies = [("game", "0047_level_70_is_unsolveable")]

    operations = [
        migrations.AddField(
            model_name="episode",
            name="r_cows",
            field=models.BooleanField(default=False),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="level",
            name="cows",
            field=models.TextField(default="[]", max_length=10000),
            preserve_default=True,
        ),
        migrations.RunPython(add_cows_block),
    ]



from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [("game", "0042_level_score_73")]

    operations = [
        migrations.RemoveField(model_name="episode", name="first_level"),
        migrations.AlterField(
            model_name="level",
            name="episode",
            field=models.ForeignKey(
                default=None,
                blank=True,
                to="game.Episode",
                null=True,
                on_delete=models.PROTECT,
            ),
        ),
    ]



from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import create_organisation_directly, join_teacher_to_organisation
from common.tests.utils.student import create_school_student_directly
from common.tests.utils.teacher import signup_teacher_directly
from deploy import captcha
from django.test.client import Client
from django.test.testcases import TestCase
from django.urls import reverse

from .utils.level import create_save_level


class LevelModerationTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.orig_captcha_enabled = captcha.CAPTCHA_ENABLED
        captcha.CAPTCHA_ENABLED = False
        super(LevelModerationTestCase, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        captcha.CAPTCHA_ENABLED = cls.orig_captcha_enabled
        super(LevelModerationTestCase, cls).tearDownClass()

    def setUp(self):
        self.client = Client()

    def test_moderation_teachers_class(self):
        level_name = "test_level1"
        email, password = signup_teacher_directly()
        _, class_name, access_code = create_class_directly(email)

        _, _, student = create_school_student_directly(access_code)

        level = create_save_level(student, level_name)

        self.teacher_login(email, password)

        level_moderation_url = reverse("level_moderation")
        response = self.client.get(level_moderation_url)
        assert class_name in response.content.decode()
        assert level_name in response.content.decode()

        # Test delete level
        delete_level_url = reverse("delete_level", kwargs={"levelID": level.id})
        self.client.post(delete_level_url)
        response = self.client.get(level_moderation_url)
        assert level_name not in response.content.decode()

    def test_moderation_another_class(self):
        level_name = "test_level2"
        email, password = signup_teacher_directly()

        email2, _ = signup_teacher_directly()
        _, class_name, access_code = create_class_directly(email2)

        _, _, student = create_school_student_directly(access_code)

        create_save_level(student, level_name)

        self.teacher_login(email, password)

        url = reverse("level_moderation")
        response = self.client.get(url)
        assert class_name not in response.content.decode()
        assert level_name not in response.content.decode()

    def test_moderation_empty_class_filter(self):
        level_name = "test_level1"
        email, password = signup_teacher_directly()
        _, class_name, access_code = create_class_directly(email)

        _, _, student = create_school_student_directly(access_code)

        create_save_level(student, level_name)

        self.teacher_login(email, password)

        url = reverse("level_moderation")
        response = self.client.post(url, {"classes": []})
        assert class_name in response.content.decode()
        assert level_name not in response.content.decode()
        assert "No levels found." in response.content.decode()

    def test_moderation_shared_with(self):
        level_name = "test_level1"
        email, password = signup_teacher_directly()
        _, class_name, access_code = create_class_directly(email)

        _, _, student = create_school_student_directly(access_code)
        _, _, student2 = create_school_student_directly(access_code)

        create_save_level(student, level_name, shared_with=[student2.new_user])

        self.teacher_login(email, password)

        url = reverse("level_moderation")
        response = self.client.get(url)
        assert class_name in response.content.decode()
        assert level_name in response.content.decode()
        assert student2.new_user.first_name in response.content.decode()

    def test_moderation_admin(self):
        level_moderation_url = reverse("level_moderation")
        level1_name = "test_level_student1"
        level2_name = "test_level_student2"

        # Create 2 teachers in the same school, one admin, one standard
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()
        school = create_organisation_directly(email1)
        join_teacher_to_organisation(email2, school.name)

        # Create one class and student for each teacher
        _, class_name1, access_code1 = create_class_directly(email1)
        _, class_name2, access_code2 = create_class_directly(email2)
        _, _, student1 = create_school_student_directly(access_code1)
        _, _, student2 = create_school_student_directly(access_code2)

        # Create one level for each student
        level1 = create_save_level(student1, level1_name)
        level2 = create_save_level(student2, level2_name)

        # Log in as teacher2, check they cannot see student1's level and cannot delete it
        self.teacher_login(email2, password2)
        response = self.client.get(level_moderation_url)
        assert class_name1 not in response.content.decode()
        assert class_name2 in response.content.decode()
        assert level1_name not in response.content.decode()
        assert level2_name in response.content.decode()
        assert student1.new_user.first_name not in response.content.decode()
        assert student2.new_user.first_name in response.content.decode()
        # Try to delete level1, it shouldn't work
        delete_level1_url = reverse("delete_level_for_editor", kwargs={"levelId": level1.id})
        response = self.client.get(delete_level1_url)
        assert response.status_code == 401
        # Check level2 is still there
        response = self.client.get(level_moderation_url)
        assert level2_name in response.content.decode()
        self.client.logout()

        # Log in as teacher1, check they can see student2's level and can delete it
        self.teacher_login(email1, password1)
        response = self.client.get(level_moderation_url)
        assert class_name1 in response.content.decode()
        assert class_name2 in response.content.decode()
        assert level1_name in response.content.decode()
        assert level2_name in response.content.decode()
        assert student1.new_user.first_name in response.content.decode()
        assert student2.new_user.first_name in response.content.decode()
        # Delete level2
        delete_level2_url = reverse("delete_level_for_editor", kwargs={"levelId": level2.id})
        response = self.client.get(delete_level2_url)
        assert response.status_code == 200
        # Check level1 is still there and level2 is not there anymore
        response = self.client.get(level_moderation_url)
        assert level1_name in response.content.decode()
        assert level2_name not in response.content.decode()
        self.client.logout()

    def teacher_login(self, email, password):
        self.client.post(
            reverse("teacher_login"),
            {"auth-username": email, "auth-password": password, "teacher_login_view-current_step": "auth"},
            follow=True,
        )



# Generated by Django 3.2.16 on 2022-12-01 10:49
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    def add_next_level_to_last_levels(apps, schema_editor):
        Level = apps.get_model("game", "Level")
        db_alias = schema_editor.connection.alias
        Level_Object_Wrapper = Level.objects.using(db_alias)
        levels = Level_Object_Wrapper.filter(
            episode__isnull=False,
            next_level=None,
        ).exclude(name="109")
        for level in levels:
            level.next_level = Level_Object_Wrapper.filter(
                name=int(level.name) + 1, episode__isnull=False
            )[0]
            level.save()

    def remove_next_level_to_last_levels(apps, schema_editor):
        Level = apps.get_model("game", "Level")
        db_alias = schema_editor.connection.alias
        Level_Object_Wrapper = Level.objects.using(db_alias)
        levels = Level_Object_Wrapper.filter(
            episode__isnull=False,
            name__in=["12", "18", "28", "32", "43", "50", "60", "67", "79", "91"],
        )
        for level in levels:
            level.next_level = None
            level.save()

    dependencies = [
        ("game", "0076_level_locked_for_class"),
    ]

    operations = [
        migrations.AlterField(
            model_name="level",
            name="next_level",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="prev_level",
                to="game.level",
            ),
        ),
        migrations.RunPython(
            add_next_level_to_last_levels, reverse_code=remove_next_level_to_last_levels
        ),
    ]



import os

from selenium import webdriver

headless_chrome_options = webdriver.ChromeOptions()
headless_chrome_options.add_argument("--headless")
headless_chrome_options.add_argument("--no-sandbox")

SELENIUM_WEBDRIVERS = {
    "default": {"callable": webdriver.Chrome, "args": (), "kwargs": {}},
    "firefox": {"callable": webdriver.Firefox, "args": (), "kwargs": {}},
    "chrome-headless": {"callable": webdriver.Chrome, "args": (), "kwargs": {"options": headless_chrome_options}},
}

SELENIUM_WIDTHS = [1624]

DEBUG = True

BASE_DIR = os.path.dirname(os.path.dirname(__file__))

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "debug": DEBUG,
            "context_processors": [
                "django.template.context_processors.request",
                "django.template.context_processors.debug",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "sekizai.context_processors.sekizai",
            ],
        },
    }
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",  # Add 'postgresql_psycopg2', 'mysql', 'sqlite3' or 'oracle'.
        "NAME": os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "db.sqlite3"
        ),  # Or path to database file if using sqlite3.
    }
}

EMAIL_BACKEND = "django.core.mail.backends.console.EmailBackend"

USE_I18N = True
USE_L10N = True
USE_TZ = True

TIME_ZONE = "Europe/London"
LANGUAGE_CODE = "en-gb"
LANGUAGES = (("en-gb", "English"), ("lol-us", "Localisation"))

LOCALE_PATHS = [
    # This shouldn't be needed, but it looks like there's an issue with
    # using a language code that's not in `django/conf/locale` - the
    # check_for_language function doesn't recognise it.
    os.path.join(os.path.dirname(__file__), "locale")
]

STATIC_ROOT = os.path.join(BASE_DIR, "static")
STATIC_URL = "/static/"
STATICFILES_DIRS = [os.path.join(BASE_DIR, "game/static")]

SECRET_KEY = "bad_test_secret"
ROOT_URLCONF = "example_project.urls"

WSGI_APPLICATION = "wsgi.application"

INSTALLED_APPS = (
    "game",
    "pipeline",
    "portal",
    "aimmo",
    "common",
    "django.contrib.admin",
    "django.contrib.admindocs",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_js_reverse",
    "django_otp",
    "django_otp.plugins.otp_static",
    "django_otp.plugins.otp_totp",
    "rest_framework",
    "import_export",
    "sekizai",  # for javascript and css management
)

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

ALLOWED_HOSTS = ["*"]
PIPELINE_ENABLED = False

PIPELINE = {
    "SASS_ARGUMENTS": "--quiet",
    "COMPILERS": ("game.pipeline_compilers.LibSassCompiler",),
    "STYLESHEETS": {
        "css": {
            "source_filenames": (
                os.path.join(BASE_DIR, "static/portal/sass/bootstrap.scss"),
                os.path.join(BASE_DIR, "static/portal/sass/colorbox.scss"),
                os.path.join(BASE_DIR, "static/portal/sass/styles.scss"),
            ),
            "output_filename": "portal.css",
        },
        "popup": {
            "source_filenames": (os.path.join(BASE_DIR, "static/portal/sass/partials/_popup.scss"),),
            "output_filename": "popup.css",
        },
        "game-scss": {
            "source_filenames": (os.path.join(BASE_DIR, "static/game/sass/game.scss"),),
            "output_filename": "game.css",
        },
    },
    "CSS_COMPRESSOR": None,
}

STATICFILES_FINDERS = [
    "pipeline.finders.PipelineFinder",
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
]
STATICFILES_STORAGE = "pipeline.storage.PipelineStorage"

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

# This is used in common to enable/disable the OneTrust cookie management script
COOKIE_MANAGEMENT_ENABLED = False

CLOUD_STORAGE_PREFIX = "https://storage.googleapis.com/codeforlife-assets/"
SITE_ID = 1

try:
    from example_project.local_settings import *  # pylint: disable=E0611
except ImportError:
    pass

from common.csp_config import *



from game.end_to_end_tests.base_game_test import BaseGameTest
from game.models import Level, Block, LevelBlock

from game.theme import get_theme
from game.character import get_character


class TestCowCrashes(BaseGameTest):
    cow_level = None

    def test_crash_into_cow_going_forward(self):
        self.run_cow_crashing_test("crash_into_cow_going_forward")

    def test_crash_into_cow_turning_left(self):
        self.run_cow_crashing_test("crash_into_cow_turning_left")

    def test_crash_into_cow_turning_right(self):
        self.run_cow_crashing_test("crash_into_cow_turning_right")

    def test_crash_into_cow_on_t_junction(self):
        self.run_cow_crashing_test("crash_into_cow_on_t_junction")

    def test_crash_into_cow_on_t_junction2(self):
        self.run_cow_crashing_test("crash_into_cow_on_t_junction2")

    def test_crash_into_cow_on_t_junction3(self):
        self.run_cow_crashing_test("crash_into_cow_on_t_junction3")

    def test_crash_into_cow_on_crossroads_junction(self):
        self.run_cow_crashing_test("crash_into_cow_on_crossroads_junction")

    def run_cow_crashing_test(self, workspace_file):
        user_profile = self.login_once()
        TestCowCrashes.cow_level.owner = user_profile
        TestCowCrashes.cow_level.save()
        workspace_id = self.use_workspace(workspace_file, user_profile)
        self.go_to_custom_level(TestCowCrashes.cow_level).load_solution(
            workspace_id
        ).run_cow_crashing_program()

    @classmethod
    def setUpClass(cls):
        BaseGameTest.setUpClass()
        grass = get_theme(name="grass")

        van = get_character("Van")

        TestCowCrashes.cow_level = Level(
            name="Cow crashing",
            anonymous=False,
            blocklyEnabled=True,
            character=van,
            cows='[{"minCows":"7","maxCows":"7","potentialCoordinates":[{"x":4,"y":4},{"x":2,"y":4},{"x":3,"y":7},{"x":4,"y":6},{"x":2,"y":6},{"x":3,"y":1},{"x":4,"y":2}],"type":"WHITE"}]',
            default=False,
            destinations="[[4,5]]",
            direct_drive=True,
            fuel_gauge=False,
            max_fuel=50,
            model_solution="[1]",
            origin='{"coordinate":[2,5],"direction":"E"}',
            path='[{"coordinate":[2,5],"connectedNodes":[1]},{"coordinate":[3,5],"connectedNodes":[0,4,2,5]},{"coordinate":[4,5],"connectedNodes":[1]},{"coordinate":[3,7],"connectedNodes":[4]},{"coordinate":[3,6],"connectedNodes":[8,3,6,1]},{"coordinate":[3,4],"connectedNodes":[10,1,11,16]},{"coordinate":[4,6],"connectedNodes":[4,7]},{"coordinate":[4,7],"connectedNodes":[6]},{"coordinate":[2,6],"connectedNodes":[9,4]},{"coordinate":[2,7],"connectedNodes":[8]},{"coordinate":[2,4],"connectedNodes":[13,5,12]},{"coordinate":[4,4],"connectedNodes":[5,14,15]},{"coordinate":[2,3],"connectedNodes":[10]},{"coordinate":[1,4],"connectedNodes":[10]},{"coordinate":[5,4],"connectedNodes":[11]},{"coordinate":[4,3],"connectedNodes":[11,19]},{"coordinate":[3,3],"connectedNodes":[5,17]},{"coordinate":[3,2],"connectedNodes":[18,16,19,20]},{"coordinate":[2,2],"connectedNodes":[17]},{"coordinate":[4,2],"connectedNodes":[17,15,23,22]},{"coordinate":[3,1],"connectedNodes":[21,17,22]},{"coordinate":[2,1],"connectedNodes":[20]},{"coordinate":[4,1],"connectedNodes":[20,19]},{"coordinate":[5,2],"connectedNodes":[19]}]',
            pythonEnabled=False,
            theme=grass,
            threads=1,
            traffic_lights="[]",
        )

        TestCowCrashes.cow_level.save()

        blocks = Block.objects.filter(
            type__in=["move_forwards", "turn_left", "turn_right"]
        )

        for block in blocks:
            new_block = LevelBlock(
                type=block, number=None, level=TestCowCrashes.cow_level
            )
            new_block.save()



# Generated by Django 3.2.18 on 2023-02-28 14:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('game', '0083_add_cows_to_existing_levels'),
    ]

    operations = [
        migrations.AlterField(
            model_name='block',
            name='block_type',
            field=models.IntegerField(choices=[(0, 'Start'), (1, 'Action'), (2, 'Condition'), (3, 'Procedure'), (4, 'ControlFlow'), (5, 'Variable'), (6, 'Math')]),
        ),
    ]



from django.db import migrations


def update_level(apps, schema_editor):
    Level = apps.get_model("game", "Level")
    level48 = Level.objects.get(name="48", default=1)
    level48.destinations = "[[1, 2], [6, 2], [6, 7], [7, 4]]"
    level48.model_solution = "[14, 15]"
    level48.save()


def revert_level(apps, schema_editor):
    Level = apps.get_model("game", "Level")
    level48 = Level.objects.get(name="48", default=1)
    level48.destinations = "[[1, 2]]"
    level48.model_solution = "[12, 13]"
    level48.save()


def add_levelblock(apps, schema_editor):
    Block = apps.get_model("game", "Block")
    Level = apps.get_model("game", "Level")
    LevelBlock = apps.get_model("game", "LevelBlock")

    level48 = Level.objects.get(name="48", default=1)
    deliver_block = Block.objects.get(type="deliver")

    newBlock = LevelBlock(type=deliver_block, number=None, level=level48)
    newBlock.save()


def remove_levelblock(apps, schema_editor):
    Block = apps.get_model("game", "Block")
    Level = apps.get_model("game", "Level")
    LevelBlock = apps.get_model("game", "LevelBlock")

    level48 = Level.objects.get(name="48", default=1)
    deliver_block = Block.objects.get(type="deliver")

    LevelBlock.objects.filter(level=level48, type=deliver_block).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("game", "0074_level_74_solution"),
    ]

    operations = [
        migrations.RunPython(update_level, reverse_code=revert_level),
        migrations.RunPython(add_levelblock, reverse_code=remove_levelblock),
    ]



# Generated by Django 3.2.16 on 2023-01-27 13:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('game', '0079_populate_block_type_add_cow_blocks'),
    ]

    operations = [
        migrations.AddField(
            model_name='level',
            name='disable_algorithm_score',
            field=models.BooleanField(default=False),
        ),
    ]



import logging

from rest_framework import permissions

LOGGER = logging.getLogger(__name__)


def _get_userprofile_school(userprofile):
    if hasattr(userprofile, "teacher"):
        return userprofile.teacher.school
    elif hasattr(userprofile, "student"):
        return userprofile.student.class_field.teacher.school
    else:
        LOGGER.error(f"Userprofile ID {userprofile.id} has no teacher or student attribute")
        return None


#########################
# Workspace permissions #
#########################


def can_create_workspace(user):
    return not user.is_anonymous


def can_load_workspace(user, workspace):
    return not user.is_anonymous and workspace.owner == user.userprofile


def can_save_workspace(user, workspace):
    return not user.is_anonymous and workspace.owner == user.userprofile


def can_delete_workspace(user, workspace):
    return not user.is_anonymous and workspace.owner == user.userprofile


#####################
# Level permissions #
#####################


def can_create_level(user):
    return not user.is_anonymous


def can_play_or_delete_level(user, level):
    # If the teacher is an admin, they can play any student's level in the school, otherwise only student levels
    # from their own classes
    if user.userprofile.teacher.is_admin and hasattr(level.owner, "student"):
        return user.userprofile.teacher.school == level.owner.student.class_field.teacher.school
    else:
        return user.userprofile.teacher.teaches(level.owner)


def can_play_level(user, level, early_access):
    if not user.is_anonymous and hasattr(user.userprofile, "student") and user.userprofile.student.class_field:
        # If the user is a student, check that the level isn't locked for their class
        return user.userprofile.student.class_field not in level.locked_for_class.all()
    elif level.default and not level.episode.in_development:
        return True
    elif level.anonymous:
        return False
    elif level.default and level.episode.in_development and early_access:
        return True
    elif user.is_anonymous:
        return level.default and not level.episode.in_development
    elif user.userprofile == level.owner:
        return True
    elif level.shared_with.filter(id=user.id):
        user_school = _get_userprofile_school(user.userprofile)
        owner_school = _get_userprofile_school(level.owner)
        return user_school is not None and user_school == owner_school
    else:
        return can_play_or_delete_level(user, level)


def can_load_level(user, level):
    if user.is_anonymous:
        return False
    elif user.userprofile == level.owner:
        return True
    elif level.shared_with.filter(id=user.id):
        user_school = _get_userprofile_school(user.userprofile)
        owner_school = _get_userprofile_school(level.owner)
        return user_school is not None and user_school == owner_school
    else:
        return hasattr(user.userprofile, "teacher") and user.userprofile.teacher.teaches(level.owner)


def can_save_level(user, level):
    if level.anonymous:
        return True
    elif user.is_anonymous:
        return False
    else:
        return user.userprofile == level.owner


def can_delete_level(user, level):
    if user.is_anonymous:
        return False
    elif level.owner == user.userprofile:
        return True
    elif hasattr(user.userprofile, "teacher"):
        return can_play_or_delete_level(user, level)

    return False


class CanShareLevel(permissions.BasePermission):
    """
    Used to verify that an incoming request is made by a user who is authorised to share
    the level - that is, that they are the owner of the level as a student, or if they're a teacher that the level was
    shared with them.
    """

    def has_permission(self, request, view):
        return True

    def has_object_permission(self, request, view, obj):
        if request.user.is_anonymous:
            return False
        elif hasattr(request.user.userprofile, "student") and request.user.userprofile.student.is_independent():
            return False
        # if the user is a teacher and the level is shared with them
        elif hasattr(request.user.userprofile, "teacher") and obj.shared_with.filter(id=request.user.id).exists():
            return True
        else:
            return obj.owner == request.user.userprofile


class CanShareLevelWith(permissions.BasePermission):
    """
    Used to verify that the user who is requesting to share their level is authorised to share the level with a specific
    recipient.
    The user is authorised if:
    - neither they nor the recipient are anonymous,
    - neither they nor the recipient are independent students,
    - they are a student and the recipient is a student in the same class, or their teacher
    - they are a teacher and the recipient is a teacher in the same school, or their student
    - they are an admin teacher and the recipient is a student in the same school
    """

    def has_permission(self, request, view):
        return True

    def has_object_permission(self, request, view, obj):
        sharer = request.user
        return self.can_share_level_with(obj, sharer)

    def can_share_level_with(self, recipient, sharer):
        if recipient.is_anonymous or sharer.is_anonymous:
            return False

        recipient_profile = recipient.userprofile
        sharer_profile = sharer.userprofile

        if (
            hasattr(sharer_profile, "student")
            and not (sharer_profile.student.is_independent())
            and hasattr(recipient_profile, "student")
            and not (recipient_profile.student.is_independent())
        ):
            # Are they in the same class?
            return sharer_profile.student.class_field == recipient_profile.student.class_field
        elif hasattr(sharer_profile, "teacher") and sharer_profile.teacher.teaches(recipient_profile):
            # Is the recipient taught by the sharer?
            return True
        elif hasattr(recipient_profile, "teacher") and recipient_profile.teacher.teaches(sharer_profile):
            # Is the sharer taught by the recipient?
            return True
        elif hasattr(sharer_profile, "teacher") and hasattr(recipient_profile, "teacher"):
            # Are they in the same organisation?
            return recipient_profile.teacher.school == sharer_profile.teacher.school
        elif hasattr(sharer_profile, "teacher") and sharer_profile.teacher.is_admin:
            return recipient_profile.student.class_field.teacher.school == sharer_profile.teacher.school
        else:
            return False


#####################
# Other permissions #
#####################


def can_see_class(user, class_):
    if user.is_anonymous:
        return False
    elif hasattr(user.userprofile, "teacher"):
        return class_.teacher == user.userprofile.teacher


def can_see_level_moderation(user):
    if user.is_anonymous:
        return False
    else:
        return hasattr(user.userprofile, "teacher")


def can_see_scoreboard(user):
    return not user.is_anonymous



from __future__ import unicode_literals

from django.db import models, migrations


def disable_route_scores(apps, schema_editor):

    Level = apps.get_model("game", "Level")
    level52 = Level.objects.get(name="52")
    level54 = Level.objects.get(name="54")
    level59 = Level.objects.get(name="59")
    level60 = Level.objects.get(name="60")
    level68 = Level.objects.get(name="68")

    level52.disable_route_score = True
    level52.save()
    level54.disable_route_score = True
    level54.save()
    level59.disable_route_score = True
    level59.save()
    level60.disable_route_score = True
    level60.save()
    level68.disable_route_score = True
    level68.save()


class Migration(migrations.Migration):

    dependencies = [("game", "0028_level_disable_route_score")]

    operations = [migrations.RunPython(disable_route_scores)]



from __future__ import absolute_import
from builtins import str
from django.core.cache import cache
from django.db.models.signals import post_save, post_delete
from django.shortcuts import get_object_or_404
from django.dispatch import receiver
from .models import Level, Episode, LevelDecor, LevelBlock
from . import level_management


LEVEL_PREFIX = "model_level"
EPISODE_PREFIX = "model_episode"
LEVEL_DECOR_PREFIX = "level_decor"
LEVEL_BLOCKS_PREFIX = "level_blocks"


def get_level(level):
    return get_object_or_404(Level, name=level, default=True)


def get_custom_level(level):
    return get_object_or_404(Level, id=level)


def get_episode(episode):
    return get_object_or_404(Episode, id=episode)


@receiver(post_save, sender=Level)
@receiver(post_delete, sender=Level)
def clear_level_cache(sender, instance, **kwargs):
    cache.delete(LEVEL_PREFIX + str(instance.id))


@receiver(post_save, sender=LevelDecor)
@receiver(post_delete, sender=LevelDecor)
def clear_level_decor_cache(sender, instance, **kwargs):
    cache.delete(LEVEL_DECOR_PREFIX + str(instance.level.id))


@receiver(post_save, sender=LevelBlock)
@receiver(post_delete, sender=LevelBlock)
def clear_level_blocks_cache(sender, instance, **kwargs):
    cache.delete(LEVEL_BLOCKS_PREFIX + str(instance.level.id))


def cache_or_func(key, func):
    result = cache.get(key)
    if result is None:
        result = func()
        cache.set(key, result)
    return result


def cached_default_level(name):
    key = LEVEL_PREFIX + str(name)
    func = lambda: get_level(name)
    return cache_or_func(key, func)


def cached_custom_level(level_id):
    key = LEVEL_PREFIX + str(level_id)
    func = lambda: get_custom_level(level_id)
    return cache_or_func(key, func)


def cached_episode(episode):
    key = EPISODE_PREFIX + episode
    func = lambda: get_episode(episode)
    return cache_or_func(key, func)


def cached_level_decor(level):
    key = LEVEL_DECOR_PREFIX + str(level.id)
    func = lambda: level_management.get_decor(level)
    return cache_or_func(key, func)


def cached_level_blocks(level):
    key = LEVEL_BLOCKS_PREFIX + str(level.id)
    func = lambda: level_management.get_blocks(level)
    return cache_or_func(key, func)



from django.db import migrations


def update_level(apps, schema_editor):
    Level = apps.get_model("game", "Level")
    level70 = Level.objects.get(name="70", default=1)
    level70.traffic_lights = (
        '[{"redDuration":3,"greenDuration":3,"sourceCoordinate":{"x":4,"y":4},"direction":"N","startTime":0,"startingState":"GREEN"},'
        '{"redDuration":3,"greenDuration":3,"sourceCoordinate":{"x":4,"y":2},"direction":"S","startTime":0,"startingState":"GREEN"},'
        '{"redDuration":3,"greenDuration":3,"sourceCoordinate":{"x":4,"y":3},"direction":"E","startTime":0,"startingState":"RED"},'
        '{"redDuration":3,"greenDuration":3,"sourceCoordinate":{"x":2,"y":3},"direction":"E","startTime":0,"startingState":"RED"}]'
    )
    level70.save()


class Migration(migrations.Migration):
    dependencies = [("game", "0052_attempt_night_mode")]

    operations = [migrations.RunPython(update_level)]
